==> squint_json-0.0.6.tar.gz: lib/src/generator/shared.dart <==
// Copyright (c) 2021 - 2022 Buijs Software
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

import "../analyzer/analyzer.dart";
import "../ast/ast.dart";
import "../common/common.dart";

/// Utilities to generate getter/setters for dataclasses and serializers.
///
/// {@category generator}
extension JsonNodeGenerator on List<TypeMember> {
  /// Generate setters.
  ///
  /// {@category generator}
  List<String> toJsonNodeSetters({String dataPrefix = ""}) =>
      map((TypeMember tm) => "    ${tm.toJsonSetter(dataPrefix: dataPrefix)},")
          .toList()
        ..sort((a, b) => a.trim().startsWith("Json") ? -1 : 1);

  /// Generate getters.
  ///
  /// {@category generator}
  List<String> toJsonNodeGetters({String dataPrefix = ""}) =>
      map((TypeMember tm) => "    ${tm.toJsonGetter(dataPrefix: dataPrefix)},")
          .toList();
}

/// Utilities to generate getters/setters for enum classes and serializers.
///
/// {@category generator}
extension JsonNodeEnumGenerator on EnumType {
  /// Generate setters.
  ///
  /// {@category generator}
  List<String> toJsonNodeSetters(String key) {
    final output = <String>[];

    var index = 0;
    while (index < values.length) {
      output.add("""
          case $className.${values[index]}:
            return const JsonString(key: "$key", data: "${valuesJSON[index]}");
          """);
      index += 1;
    }

    return output;
  }

  /// Generate getters.
  ///
  /// {@category generator}
  List<String> toJsonNodeGetters(String key) {
    final output = <String>[];

    var index = 0;
    while (index < values.length) {
      output.add("""
          case "${valuesJSON[index]}":
            return $className.${values[index]};
          """);
      index += 1;
    }

    output.add("""
        default:
          throw SquintException("Unable to map value to $className enum: \${value.data}");
        """);

    return output;
  }
}

extension on TypeMember {
  String toJsonGetter({String dataPrefix = ""}) {
    final type = this.type.className;

    final valueJsonAnnotated = annotations
        .firstBy((element) => element.name == "JsonValue")
        ?.data["tag"];

    final jsonKey = valueJsonAnnotated ?? name;

    final decodeJsonAnnotated =
        annotations.firstBy((element) => element.name == "JsonDecode");

    final q = this.type.nullable ? "OrNull" : "";

    if (decodeJsonAnnotated != null) {
      final method = decodeJsonAnnotated.data["using"];

      final unwrapperType = decodeJsonAnnotated.data["jsonElement"];

      String? unwrapper;

      if (unwrapperType == "JsonString") {
        unwrapper = '${dataPrefix}stringNode$q("$jsonKey")';
      } else if (unwrapperType == "JsonFloatingNumber") {
        unwrapper = '${dataPrefix}floatNode$q("$jsonKey")';
      } else if (unwrapperType == "JsonIntegerNumber") {
        unwrapper = '${dataPrefix}integerNode$q("$jsonKey")';
      } else if (unwrapperType == "JsonBoolean") {
        unwrapper = '${dataPrefix}booleanNode$q("$jsonKey")';
      } else if (unwrapperType == "JsonArray") {
        unwrapper =
            '${dataPrefix}arrayNode$q<${(this.type as ListType).child.printType}>("$jsonKey")';
      } else if (unwrapperType == "JsonObject") {
        unwrapper = '${dataPrefix}objectNode$q("$jsonKey")';
      } else {
        throw SquintException("Unsupported data type: $unwrapperType");
      }

      return "$name: $method($unwrapper)";
    }

    switch (type) {
      case "String":
        return '$name: ${dataPrefix}string$q("$jsonKey")';
      case "double":
        return '$name: ${dataPrefix}float$q("$jsonKey")';
      case "int":
        return '$name: ${dataPrefix}integer$q("$jsonKey")';
      case "bool":
        return '$name: ${dataPrefix}boolean$q("$jsonKey")';
      case "List":
        return '$name: ${dataPrefix}array$q<${(this.type as ListType).child.printType}>("$jsonKey")';
      case "Map":
        return '$name: ${dataPrefix}object$q("$jsonKey")';
      case "dynamic":
        return '$name: ${dataPrefix}byKey("$jsonKey").data';
      default:
        throw SquintException("Unsupported data type: $type");
    }
  }

  String toJsonSetter({String dataPrefix = ""}) {
    final valueJsonAnnotated = annotations
        .firstBy((element) => element.name == "JsonValue")
        ?.data["tag"];

    final jsonKey = valueJsonAnnotated ?? name;

    final encodeJsonAnnotated = annotations
        .firstBy((element) => element.name == "JsonEncode")
        ?.data["using"];

    if (encodeJsonAnnotated != null) {
      return "$encodeJsonAnnotated($dataPrefix$name)";
    }

    final q = type.nullable ? "OrNull" : "";

    switch (type.className) {
      case "String":
        return 'JsonString$q(key: "$jsonKey", data: $dataPrefix$name)';
      case "double":
        return 'JsonFloatingNumber$q(key: "$jsonKey", data: $dataPrefix$name)';
      case "int":
        return 'JsonIntegerNumber$q(key: "$jsonKey", data: $dataPrefix$name)';
      case "bool":
        return 'JsonBoolean$q(key: "$jsonKey", data: $dataPrefix$name)';
      case "List":
        return 'JsonArray$q<dynamic>(key: "$jsonKey", data: $dataPrefix$name)';
      case "Map":
        return 'JsonObject$q.fromMap(key: "$jsonKey", data: $dataPrefix$name)';
      case "dynamic":
        return 'dynamicValue(key: "$jsonKey", data: $dataPrefix$name)';
      default:
        throw SquintException("Unsupported data type: ${type.className}");
    }
  }
}

==> squint_json-0.1.1.tar.gz: lib/src/generator/shared.dart <==
// Copyright (c) 2021 - 2022 Buijs Software
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

import "../analyzer/analyzer.dart";
import "../ast/ast.dart";
import "../common/common.dart";

/// Utilities to generate getter/setters for dataclasses and serializers.
///
/// {@category generator}
extension JsonNodeGenerator on List<TypeMember> {
  /// Generate setters.
  ///
  /// {@category generator}
  List<String> toJsonNodeSetters({String dataPrefix = ""}) =>
      map((TypeMember tm) => "    ${tm.toJsonSetter(dataPrefix: dataPrefix)},")
          .toList()
        ..sort((a, b) => a.trim().startsWith("Json") ? -1 : 1);

  /// Generate getters.
  ///
  /// {@category generator}
  List<String> toJsonNodeGetters({String dataPrefix = ""}) =>
      map((TypeMember tm) => "    ${tm.toJsonGetter(dataPrefix: dataPrefix)},")
          .toList();
}

/// Utilities to generate getters/setters for enum classes and serializers.
///
/// {@category generator}
extension JsonNodeEnumGenerator on EnumType {
  /// Generate setters.
  ///
  /// {@category generator}
  List<String> toJsonNodeSetters(String key) {
    final output = <String>[];

    var index = 0;
    while (index < values.length) {
      output.add("""
          case $className.${values[index]}:
            return const JsonString(key: "$key", data: "${valuesJSON[index]}");
          """);
      index += 1;
    }

    return output;
  }

  /// Generate getters.
  ///
  /// {@category generator}
  List<String> toJsonNodeGetters(String key) {
    final output = <String>[];

    var index = 0;
    while (index < values.length) {
      output.add("""
          case "${valuesJSON[index]}":
            return $className.${values[index]};
          """);
      index += 1;
    }

    return output;
  }
}

extension on TypeMember {
  String toJsonGetter({String dataPrefix = ""}) {
    final className = type.className;

    final valueJsonAnnotated = annotations
        .firstBy((element) => element.name == "JsonValue")
        ?.data["tag"];

    final jsonKey = valueJsonAnnotated ?? name;

    final decodeJsonAnnotated =
        annotations.firstBy((element) => element.name == "JsonDecode");

    final q = type.nullable ? "OrNull" : "";

    if (decodeJsonAnnotated != null) {
      final method = decodeJsonAnnotated.data["using"];

      final unwrapperType = decodeJsonAnnotated.data["jsonElement"];

      String? unwrapper;

      if (unwrapperType == "JsonString") {
        unwrapper = '${dataPrefix}stringNode$q("$jsonKey")';
      } else if (unwrapperType == "JsonFloatingNumber") {
        unwrapper = '${dataPrefix}floatNode$q("$jsonKey")';
      } else if (unwrapperType == "JsonIntegerNumber") {
        unwrapper = '${dataPrefix}integerNode$q("$jsonKey")';
      } else if (unwrapperType == "JsonBoolean") {
        unwrapper = '${dataPrefix}booleanNode$q("$jsonKey")';
      } else if (unwrapperType == "JsonArray") {
        unwrapper =
            '${dataPrefix}arrayNode$q<${(type as ListType).child.printType}>("$jsonKey")';
      } else if (unwrapperType == "JsonObject") {
        unwrapper = '${dataPrefix}objectNode$q("$jsonKey")';
      } else {
        throw SquintException("Unsupported data type: $unwrapperType");
      }

      return "$name: $method($unwrapper)";
    }

    switch (className) {
      case "String":
        return '$name: ${dataPrefix}string$q("$jsonKey")';
      case "double":
        return '$name: ${dataPrefix}float$q("$jsonKey")';
      case "int":
        return '$name: ${dataPrefix}integer$q("$jsonKey")';
      case "bool":
        return '$name: ${dataPrefix}boolean$q("$jsonKey")';
      case "List":
        return '$name: ${dataPrefix}array$q<${(type as ListType).child.printType}>("$jsonKey")';
      case "Map":
        final mapType = type as MapType;
        final mapKeyType = mapType.key;
        final mapKeyTypeString = mapKeyType.className;
        if (mapKeyType is EnumType) {
          return '$name: ${dataPrefix}enumObject<$mapKeyTypeString,${mapType.value.className}>$q(key: "$jsonKey", keyToEnumValue: (String entry) => $mapKeyTypeString.values.firstWhere((value) => value == entry, orElse: () => $mapKeyTypeString.${mapKeyType.noneValue}))';
        } else {
          return '$name: ${dataPrefix}object$q("$jsonKey")';
        }
      case "dynamic":
        return '$name: ${dataPrefix}byKey("$jsonKey").data';
      default:
        throw SquintException("Unsupported data type: $className");
    }
  }

  String toJsonSetter({String dataPrefix = ""}) {
    final valueJsonAnnotated = annotations
        .firstBy((element) => element.name == "JsonValue")
        ?.data["tag"];

    final jsonKey = valueJsonAnnotated ?? name;

    final encodeJsonAnnotated = annotations
        .firstBy((element) => element.name == "JsonEncode")
        ?.data["using"];

    if (encodeJsonAnnotated != null) {
      return "$encodeJsonAnnotated($dataPrefix$name)";
    }

    final q = type.nullable ? "OrNull" : "";

    switch (type.className) {
      case "String":
        return 'JsonString$q(key: "$jsonKey", data: $dataPrefix$name)';
      case "double":
        return 'JsonFloatingNumber$q(key: "$jsonKey", data: $dataPrefix$name)';
      case "int":
        return 'JsonIntegerNumber$q(key: "$jsonKey", data: $dataPrefix$name)';
      case "bool":
        return 'JsonBoolean$q(key: "$jsonKey", data: $dataPrefix$name)';
      case "List":
        return 'JsonArray$q<dynamic>(key: "$jsonKey", data: $dataPrefix$name)';
      case "Map":
        final mapType = type as MapType;
        final mapKeyType = mapType.key;
        final mapKeyTypeString = mapKeyType.className;
        if (mapKeyType is EnumType) {
          return 'JsonObject$q.fromEnumMap<$mapKeyTypeString>(keyToString: ($mapKeyTypeString entry) => entry.name, key: "$jsonKey", data: $dataPrefix$name)';
        } else {
          return 'JsonObject$q.fromMap(key: "$jsonKey", data: $dataPrefix$name)';
        }
      case "dynamic":
        return 'dynamicValue(key: "$jsonKey", data: $dataPrefix$name)';
      default:
        throw SquintException("Unsupported data type: ${type.className}");
    }
  }
}

/// Print import statements.
extension ImportsBuilder on CustomType {
  /// Return import statement for all non-standard types.
  Set<String> importStatements(Set<AbstractType> types) {
    types
      ..removeWhere((type) => type is StandardType)
      ..removeWhere((type) => type.className == className);
    return types
        .map((e) => e.className.snakeCase)
        .map((e) => "import '${e}_dataclass.dart';\n")
        .toSet();
  }
}

/// Return [String] enum value for empty JSON String.
extension NoneValueBuilder on EnumType {
  /// Return "NONE" if all enum values are uppercase or "none" if not.
  String get noneValue =>
      values.every((e) => e == e.toUpperCase()) ? "NONE" : "none";
}

/// Utility to normalize user defined Types.
extension AbstractTypeNormalizer on AbstractType {
  /// Normalize [AbstractType] by replacing any (nested Type)
  /// with the instance scanned by the analyzer.
  AbstractType normalizeType(
    Set<EnumType> enumTypes,
    Set<CustomType> customTypes,
  ) {
    final customTypeOrNull =
        customTypes.firstBy((type) => type.className == className);
    if (customTypeOrNull != null) {
      return customTypeOrNull;
    }

    final enumTypeOrNull =
        enumTypes.firstBy((type) => type.className == className);
    if (enumTypeOrNull != null) {
      return enumTypeOrNull;
    }

    if (this is ListType) {
      final listType = this as ListType;
      final childType = listType.child.normalizeType(enumTypes, customTypes);
      return listType.nullable
          ? NullableListType(childType)
          : ListType(childType);
    }

    if (this is MapType) {
      final mapType = this as MapType;
      final keyType = mapType.key.normalizeType(enumTypes, customTypes);
      final valueType = mapType.value.normalizeType(enumTypes, customTypes);
      return mapType.nullable
          ? NullableMapType(key: keyType, value: valueType)
          : MapType(key: keyType, value: valueType);
    }

    return this;
  }
}
